@@ -1,18 +1,23 @@
 #include "source-copy.hpp"
 #include <obs-module.h>
 #include <QClipboard>
+#include <QDesktopServices>
 #include <QFileDialog>
 #include <QGuiApplication>
 #include <QLabel>
+#include <QLineEdit>
 #include <QMainWindow>
 #include <QMenu>
 #include <QWidgetAction>
-#include <QLineEdit>
 
-#include "version.h"
+#include "obs-websocket-api.h"
 #include "util/config-file.h"
 #include "util/platform.h"
-#include "obs-websocket-api.h"
+#include "version.h"
+
+#ifndef _WIN32
+#include <dlfcn.h>
+#endif
 
 #define QT_UTF8(str) QString::fromUtf8(str)
 #define QT_TO_UTF8(str) str.toUtf8().constData()
@@ -133,7 +138,7 @@ static void try_fix_paths(obs_data_t *data, QString fileName)
 
 static void LoadSourceMenu(QMenu *menu, obs_source_t *source, obs_sceneitem_t *item);
 
-static void LoadSources(obs_data_array_t *data, obs_scene_t *scene)
+static void LoadSources(obs_data_array_t *data, obs_scene_t *scene, obs_canvas_t *canvas = nullptr)
 {
 	const size_t count = obs_data_array_count(data);
 	std::vector<obs_source_t *> sources;
@@ -141,8 +146,23 @@ static void LoadSources(obs_data_array_t *data, obs_scene_t *scene)
 	for (size_t i = 0; i < count; i++) {
 		obs_data_t *sourceData = obs_data_array_item(data, i);
 		const char *name = obs_data_get_string(sourceData, "name");
-
-		obs_source_t *s = obs_get_source_by_name(name);
+		const char *canvas_uuid = obs_data_get_string(sourceData, "canvas_uuid");
+		if (canvas && canvas_uuid && canvas_uuid[0] != '\0' && strcmp(canvas_uuid, obs_canvas_get_uuid(canvas)) != 0) {
+			obs_data_set_string(sourceData, "canvas_uuid", obs_canvas_get_uuid(canvas));
+			obs_source_t *found = obs_get_source_by_uuid(obs_data_get_string(sourceData, "uuid"));
+			if (found) {
+				obs_canvas_t *found_canvas = obs_source_get_canvas(found);
+				if (found_canvas != canvas) {
+					obs_data_unset_user_value(sourceData, "uuid");
+				}
+				obs_canvas_release(found_canvas);
+				obs_source_release(found);
+			}
+		}
+		obs_source_t *s =
+			(canvas && obs_get_source_output_flags(obs_data_get_string(sourceData, "id")) & OBS_SOURCE_REQUIRES_CANVAS)
+				? obs_canvas_get_source_by_name(canvas, name)
+				: obs_get_source_by_name(name);
 		if (!s)
 			s = obs_load_source(sourceData);
 		if (s) {
@@ -169,42 +189,25 @@ static void LoadSources(obs_data_array_t *data, obs_scene_t *scene)
 		obs_source_release(source);
 }
 
-static void LoadScene(obs_data_t *data)
+static void LoadSceneCanvas(obs_data_t *data, obs_canvas_t *canvas)
 {
 	if (!data)
 		return;
 	obs_data_array_t *sourcesData = obs_data_get_array(data, "sources");
 	if (!sourcesData)
 		return;
-	LoadSources(sourcesData, nullptr);
+	LoadSources(sourcesData, nullptr, canvas);
 	obs_data_array_release(sourcesData);
 }
 
-static config_t *(*get_user_config_func)(void) = nullptr;
+static void LoadScene(obs_data_t *data)
+{
+	LoadSceneCanvas(data, nullptr);
+}
 
-static config_t *get_user_config(void)
+config_t *get_user_config(void)
 {
-#if LIBOBS_API_VER < MAKE_SEMANTIC_VERSION(31, 0, 0)
-	if (!get_user_config_func) {
-		if (obs_get_version() < MAKE_SEMANTIC_VERSION(31, 0, 0)) {
-			get_user_config_func = obs_frontend_get_global_config;
-			blog(LOG_INFO, "[Aitum Multistream] use global config");
-		} else {
-			auto handle = os_dlopen("obs-frontend-api");
-			if (handle) {
-				get_user_config_func = (config_t * (*)(void)) os_dlsym(handle, "obs_frontend_get_user_config");
-				os_dlclose(handle);
-				if (get_user_config_func)
-					blog(LOG_INFO, "[Aitum Multistream] use user config");
-			}
-		}
-	}
-	if (get_user_config_func)
-		return get_user_config_func();
-	return obs_frontend_get_global_config();
-#else
 	return obs_frontend_get_user_config();
-#endif
 }
 
 obs_data_array_t *GetScriptsData()
@@ -255,10 +258,10 @@ void LoadScriptData(obs_data_t *script_data)
 		obs_data_array_release(scripts);
 		obs_data_save_json_safe(data, path.c_str(), "tmp", "bak");
 		obs_data_release(data);
-		config_set_string(config, "Basic", "SceneCollection", "");
+		config_set_string(config, "Basic", "SceneCollection", "Source Copy Temp");
 		config_set_string(config, "Basic", "SceneCollectionFile", "source_copy_temp");
 		obs_frontend_set_current_scene_collection(sceneCollection.c_str());
-		std::string temp_path = obs_module_config_path("../../basic/scenes/scene_collection_manager_temp.json");
+		std::string temp_path = obs_module_config_path("../../basic/scenes/source_copy_temp.json");
 		os_unlink(temp_path.c_str());
 	} else {
 		obs_data_release(data);
@@ -330,28 +333,28 @@ static void LoadScriptMenu(QMenu *menu)
 	obs_data_array_release(scripts);
 }
 
-static void LoadMenu(QMenu *menu)
+static void LoadCanvasMenu(QMenu *menu, obs_canvas_t *canvas)
 {
 	menu->clear();
 	QAction *a = menu->addAction(obs_module_text("LoadScene"));
-	QObject::connect(a, &QAction::triggered, [] {
+	QObject::connect(a, &QAction::triggered, [canvas] {
 		QString fileName = QFileDialog::getOpenFileName(nullptr, QT_UTF8(obs_module_text("LoadScene")), QString(),
 								"JSON File (*.json)");
 		if (fileName.isEmpty())
 			return;
 		obs_data_t *data = obs_data_create_from_json_file(QT_TO_UTF8(fileName));
 		try_fix_paths(data, fileName);
-		LoadScene(data);
+		LoadSceneCanvas(data, canvas);
 		obs_data_release(data);
 	});
 	a = menu->addAction(QT_UTF8(obs_module_text("PasteScene")));
-	QObject::connect(a, &QAction::triggered, [] {
+	QObject::connect(a, &QAction::triggered, [canvas] {
 		QClipboard *clipboard = QGuiApplication::clipboard();
 		const QString strData = clipboard->text();
 		if (strData.isEmpty())
 			return;
 		obs_data_t *data = obs_data_create_from_json(QT_TO_UTF8(strData));
-		LoadScene(data);
+		LoadSceneCanvas(data, canvas);
 		obs_data_release(data);
 	});
 	auto label = new QLabel("<b>" + QT_UTF8(obs_module_text("Scenes")) + "</b>");
@@ -361,8 +364,6 @@ static void LoadMenu(QMenu *menu)
 	wa->setDefaultWidget(label);
 	menu->addAction(wa);
 
-	struct obs_frontend_source_list scenes = {};
-	obs_frontend_get_scenes(&scenes);
 	wa = new QWidgetAction(menu);
 	auto t = new QLineEdit;
 	t->connect(t, &QLineEdit::textChanged, [menu](const QString text) {
@@ -370,7 +371,7 @@ static void LoadMenu(QMenu *menu)
 		{
 			if (!action->menu())
 				continue;
-			if (text.isEmpty() || action->text() == QT_UTF8(obs_module_text("Scripts"))) {
+			if (text.isEmpty()) {
 				action->setVisible(true);
 			} else {
 				action->setVisible(action->text().contains(text, Qt::CaseInsensitive));
@@ -379,19 +380,56 @@ static void LoadMenu(QMenu *menu)
 	});
 	wa->setDefaultWidget(t);
 	menu->addAction(wa);
+	if (canvas) {
+		obs_canvas_enum_scenes(
+			canvas,
+			[](void *data, obs_source_t *scene) -> bool {
+				QMenu *menu = static_cast<QMenu *>(data);
 
-	for (size_t i = 0; i < scenes.sources.num; i++) {
-		obs_source_t *source = scenes.sources.array[i];
-		QMenu *submenu = menu->addMenu(obs_source_get_name(scenes.sources.array[i]));
-		QObject::connect(submenu, &QMenu::aboutToShow, [submenu, source] { LoadSourceMenu(submenu, source, nullptr); });
+				QMenu *submenu = menu->addMenu(obs_source_get_name(scene));
+				QObject::connect(submenu, &QMenu::aboutToShow,
+						 [submenu, scene] { LoadSourceMenu(submenu, scene, nullptr); });
+				return true;
+			},
+			menu);
+	} else {
+		struct obs_frontend_source_list scenes = {};
+		obs_frontend_get_scenes(&scenes);
+		for (size_t i = 0; i < scenes.sources.num; i++) {
+			obs_source_t *source = scenes.sources.array[i];
+			QMenu *submenu = menu->addMenu(obs_source_get_name(scenes.sources.array[i]));
+			QObject::connect(submenu, &QMenu::aboutToShow,
+					 [submenu, source] { LoadSourceMenu(submenu, source, nullptr); });
+		}
+		obs_frontend_source_list_free(&scenes);
 	}
+}
 
-	obs_frontend_source_list_free(&scenes);
+static void LoadMenu(QMenu *menu)
+{
+	menu->clear();
+
+	obs_enum_canvases(
+		[](void *data, obs_canvas_t *canvas) -> bool {
+			QMenu *menu = static_cast<QMenu *>(data);
+			if (!canvas)
+				return true;
+			auto canvasMenu = menu->addMenu(QT_UTF8(obs_canvas_get_name(canvas)));
+			QObject::connect(canvasMenu, &QMenu::aboutToShow,
+					 [canvasMenu, canvas] { LoadCanvasMenu(canvasMenu, canvas); });
+			return true;
+		},
+		menu);
 
 	menu->addSeparator();
 
 	QMenu *submenu = menu->addMenu(QT_UTF8(obs_module_text("Scripts")));
 	QObject::connect(submenu, &QMenu::aboutToShow, [submenu] { LoadScriptMenu(submenu); });
+
+	menu->addSeparator();
+	menu->addAction(QString::fromUtf8("Source Copy (" PROJECT_VERSION ")"),
+			[] { QDesktopServices::openUrl(QUrl("https://obsproject.com/forum/resources/source-copy.1261/")); });
+	menu->addAction(QString::fromUtf8("By Exeldro"), [] { QDesktopServices::openUrl(QUrl("https://exeldro.com")); });
 }
 
 void CopyTransform(void *data, obs_hotkey_id id, obs_hotkey_t *hotkey, bool pressed)
@@ -579,12 +617,8 @@ static obs_data_t *GetTransformData(obs_sceneitem_t *item)
 {
 	obs_data_t *temp = obs_data_create();
 	obs_transform_info info{};
-#if LIBOBS_API_VER < MAKE_SEMANTIC_VERSION(30, 1, 0)
-	obs_sceneitem_get_info(item, &info);
-#else
 	obs_sceneitem_get_info2(item, &info);
 	obs_data_set_bool(temp, "crop_to_bounds", info.crop_to_bounds);
-#endif
 	obs_data_set_vec2(temp, "pos", &info.pos);
 	obs_data_set_vec2(temp, "scale", &info.scale);
 	obs_data_set_double(temp, "rot", info.rot);
@@ -604,12 +638,8 @@ static obs_data_t *GetTransformData(obs_sceneitem_t *item)
 void LoadTransform(obs_sceneitem_t *item, obs_data_t *data)
 {
 	obs_transform_info info{};
-#if LIBOBS_API_VER < MAKE_SEMANTIC_VERSION(30, 1, 0)
-	obs_sceneitem_get_info(item, &info);
-#else
 	obs_sceneitem_get_info2(item, &info);
 	info.crop_to_bounds = obs_data_get_bool(data, "crop_to_bounds");
-#endif
 	obs_data_get_vec2(data, "pos", &info.pos);
 	obs_data_get_vec2(data, "scale", &info.scale);
 	info.rot = obs_data_get_double(data, "rot");
@@ -617,11 +647,7 @@ void LoadTransform(obs_sceneitem_t *item, obs_data_t *data)
 	info.bounds_type = (enum obs_bounds_type)obs_data_get_int(data, "bounds_type");
 	obs_data_get_vec2(data, "bounds", &info.bounds);
 	info.bounds_alignment = obs_data_get_int(data, "bounds_alignment");
-#if LIBOBS_API_VER < MAKE_SEMANTIC_VERSION(30, 1, 0)
-	obs_sceneitem_set_info(item, &info);
-#else
 	obs_sceneitem_set_info2(item, &info);
-#endif
 	obs_sceneitem_crop crop{};
 	crop.top = obs_data_get_int(data, "top");
 	crop.bottom = obs_data_get_int(data, "bottom");
@@ -657,6 +683,7 @@ static void LoadSourceMenu(QMenu *menu, obs_source_t *source, obs_sceneitem_t *i
 			obs_data_array_push_back(sources, sceneData);
 			obs_data_release(sceneData);
 			obs_data_save_json(data, QT_TO_UTF8(fileName));
+			obs_data_array_release(sources);
 			obs_data_release(data);
 		});
 		a = menu->addAction(
@@ -671,6 +698,7 @@ static void LoadSourceMenu(QMenu *menu, obs_source_t *source, obs_sceneitem_t *i
 			obs_data_release(sceneData);
 			QClipboard *clipboard = QGuiApplication::clipboard();
 			clipboard->setText(QT_UTF8(obs_data_get_json(data)));
+			obs_data_array_release(sources);
 			obs_data_release(data);
 		});
 		a = menu->addAction(QT_UTF8(obs_module_text("LoadSource")));
@@ -925,7 +953,15 @@ static void *vendor;
 void websocket_add_scene(obs_data_t *request_data, obs_data_t *response_data, void *param)
 {
 	UNUSED_PARAMETER(param);
-	LoadScene(request_data);
+	obs_queue_task(OBS_TASK_UI, (obs_task_t)LoadScene, request_data, true);
+	obs_data_set_bool(response_data, "success", true);
+}
+
+void websocket_get_version(obs_data_t *request_data, obs_data_t *response_data, void *param)
+{
+	UNUSED_PARAMETER(param);
+	UNUSED_PARAMETER(request_data);
+	obs_data_set_string(response_data, "version", PROJECT_VERSION);
 	obs_data_set_bool(response_data, "success", true);
 }
 
@@ -933,7 +969,27 @@ void websocket_get_current_scene(obs_data_t *request_data, obs_data_t *response_
 {
 	UNUSED_PARAMETER(param);
 	UNUSED_PARAMETER(request_data);
-	obs_source_t *source = obs_frontend_get_current_scene();
+	obs_source_t *source = nullptr;
+	const char *canvas_name = obs_data_get_string(request_data, "canvas");
+	if (strlen(canvas_name)) {
+		obs_canvas_t *canvas = obs_get_canvas_by_name(canvas_name);
+		if (!canvas) {
+			obs_data_set_string(response_data, "error", "canvas not found");
+			obs_data_set_bool(response_data, "success", false);
+			return;
+		}
+		source = obs_canvas_get_channel(canvas, 0);
+		if (source && obs_source_get_type(source) == OBS_SOURCE_TYPE_TRANSITION) {
+			obs_source_t *ts = obs_transition_get_active_source(source);
+			if (ts) {
+				obs_source_release(source);
+				source = ts;
+			}
+		}
+		obs_canvas_release(canvas);
+	} else {
+		source = obs_frontend_get_current_scene();
+	}
 	if (!source) {
 		obs_data_set_bool(response_data, "success", false);
 		return;
@@ -945,6 +1001,7 @@ void websocket_get_current_scene(obs_data_t *request_data, obs_data_t *response_
 	obs_data_t *sceneData = obs_save_source(source);
 	obs_data_array_push_back(sources, sceneData);
 	obs_data_release(sceneData);
+	obs_data_array_release(sources);
 	obs_source_release(source);
 	obs_data_set_bool(response_data, "success", true);
 }
@@ -959,7 +1016,21 @@ void websocket_get_scene(obs_data_t *request_data, obs_data_t *response_data, vo
 		return;
 	}
 
-	obs_source_t *source = obs_get_source_by_name(name);
+	obs_source_t *source = nullptr;
+	const char *canvas_name = obs_data_get_string(request_data, "canvas");
+	if (strlen(canvas_name)) {
+		obs_canvas_t *canvas = obs_get_canvas_by_name(canvas_name);
+		if (!canvas) {
+			obs_data_set_string(response_data, "error", "canvas not found");
+			obs_data_set_bool(response_data, "success", false);
+			return;
+		}
+		source = obs_canvas_get_source_by_name(canvas, name);
+		obs_canvas_release(canvas);
+	} else {
+		source = obs_get_source_by_name(name);
+	}
+
 	if (!source) {
 		obs_data_set_string(response_data, "error", "scene not found");
 		obs_data_set_bool(response_data, "success", false);
@@ -978,6 +1049,7 @@ void websocket_get_scene(obs_data_t *request_data, obs_data_t *response_data, vo
 	obs_data_t *sceneData = obs_save_source(source);
 	obs_data_array_push_back(sources, sceneData);
 	obs_data_release(sceneData);
+	obs_data_array_release(sources);
 	obs_source_release(source);
 	obs_data_set_bool(response_data, "success", true);
 }
@@ -992,7 +1064,20 @@ void websocket_get_source(obs_data_t *request_data, obs_data_t *response_data, v
 		return;
 	}
 
-	obs_source_t *source = obs_get_source_by_name(name);
+	obs_source_t *source = nullptr;
+	const char *canvas_name = obs_data_get_string(request_data, "canvas");
+	if (strlen(canvas_name)) {
+		obs_canvas_t *canvas = obs_get_canvas_by_name(canvas_name);
+		if (!canvas) {
+			obs_data_set_string(response_data, "error", "canvas not found");
+			obs_data_set_bool(response_data, "success", false);
+			return;
+		}
+		source = obs_canvas_get_source_by_name(canvas, name);
+		obs_canvas_release(canvas);
+	} else {
+		source = obs_get_source_by_name(name);
+	}
 	if (!source) {
 		obs_data_set_string(response_data, "error", "source not found");
 		obs_data_set_bool(response_data, "success", false);
@@ -1011,7 +1096,19 @@ void websocket_add_source(obs_data_t *request_data, obs_data_t *response_data, v
 	obs_source_t *source = nullptr;
 	const char *name = obs_data_get_string(request_data, "scene");
 	if (!name || !strlen(name)) {
-		source = obs_get_source_by_name(name);
+		const char *canvas_name = obs_data_get_string(request_data, "canvas");
+		if (strlen(canvas_name)) {
+			obs_canvas_t *canvas = obs_get_canvas_by_name(canvas_name);
+			if (!canvas) {
+				obs_data_set_string(response_data, "error", "canvas not found");
+				obs_data_set_bool(response_data, "success", false);
+				return;
+			}
+			source = obs_canvas_get_source_by_name(canvas, name);
+			obs_canvas_release(canvas);
+		} else {
+			source = obs_get_source_by_name(name);
+		}
 	} else {
 		source = obs_frontend_get_current_scene();
 	}
@@ -1037,6 +1134,7 @@ void obs_module_post_load(void)
 	vendor = obs_websocket_register_vendor("source-copy");
 	if (!vendor)
 		return;
+	obs_websocket_vendor_register_request(vendor, "get_version", websocket_get_version, nullptr);
 	obs_websocket_vendor_register_request(vendor, "get_current_scene", websocket_get_current_scene, nullptr);
 	obs_websocket_vendor_register_request(vendor, "get_scene", websocket_get_scene, nullptr);
 	obs_websocket_vendor_register_request(vendor, "add_scene", websocket_add_scene, nullptr);
